/*
 * Very simple USB CDC driver for SAM3S
 *  This driver code is derived from ATMEL library code.
 */
#include <stdio.h>
#include <string.h>
#include "SAMCamera.h"
#include "usb.h"

enum {
  EP_IDLE, EP_SENDING, EP_RECEIVING, EP_DONE,
} epstate;

static uint16_t addr_flag;

struct usb_info {
  uint8_t currentConfiguration;
  uint8_t currentConnection;
} usb;

struct ep_desc {
  uint16_t epsize;
  uint16_t dualbank;
  uint16_t epstate;
  uint16_t remain;		/* remaining size waiting for FIFO */
  uint16_t filled;		/* size already filled into FIFO */
  uint16_t done;		/* Size of transfer finished */
  uint8_t *pdata;
  uint8_t cbank;		/* current bank number */
  void (*callback)();		/* callback handler */
};

struct ep_desc epdesctab[] = {
  { EP_SIZE_CTRL, 0, EP_IDLE },
  { EP_SIZE_BULK, 1, EP_IDLE},
  { EP_SIZE_BULK, 1, EP_IDLE },
  { EP_SIZE_BULK, 0, EP_IDLE },
};

#define	EPDESCTAB_SIZE	(sizeof(epdesctab)/sizeof(struct ep_desc))

#define	MIN(x, y)	(x > y)? y : x

#define REG_NO_EFFECT_1_ALL	UDP_CSR_RX_DATA_BK0 | UDP_CSR_RX_DATA_BK1 \
		|UDP_CSR_STALLSENTISOERROR   | UDP_CSR_RXSETUP \
		|UDP_CSR_TXCOMP

/*  Sets the specified bit(s) in the UDP_CSR register.
 *    endpoint: The endpoint number of the CSR to process.
 *    flags: The bitmap to clear to 0.
 */
#define SET_CSR(endpoint, flags) \
    { \
        volatile unsigned int reg; \
        Udp *baseudp = UDP; \
        reg = baseudp->UDP_CSR[endpoint] ; \
        reg |= REG_NO_EFFECT_1_ALL; \
        reg |= (flags); \
        baseudp->UDP_CSR[endpoint] = reg; \
        while ( (baseudp->UDP_CSR[endpoint] & (flags)) != (flags)); \
    }

/*  Clears the specified bit(s) in the UDP_CSR register.
 *    endpoint: The endpoint number of the CSR to process.
 *    flags: The bitmap to set to 1.
 */
#define CLEAR_CSR(endpoint, flags) \
    { \
        volatile unsigned int reg; \
        Udp *baseudp = UDP; \
        reg = baseudp->UDP_CSR[endpoint]; \
        reg |= REG_NO_EFFECT_1_ALL; \
        reg &= ~(flags); \
        baseudp->UDP_CSR[endpoint] = reg; \
        while ( (baseudp->UDP_CSR[endpoint] & (flags)) == (flags)); \
    }

static void usb_process_setup(Udp *pUDP, struct ep_desc *pEpDesc);

struct usb_pkt req_pkt;

const uint8_t devDescriptor[] = {	/* Device descriptor */
    0x12,	/* bLength */
    DESC_DEVICE,	/* bDescriptorType */
    0x00, 0x02,	/* bCDUSB */
    0x02,	/* bDeviceClass: CDC class code */
    0x00,	/* bDeviceSubclass: CDC class sub code */
    0x00,	/* bDeviceProtocol: CDC Device protocol */
    EP_SIZE_CTRL,	/* bMaxPacketSize0 */
    0xEB, 0x03,	/* idVendor */
    0x19, 0x61,	/* idProduct */
    0x10, 0x01,	/* bcdDevice */
    0x01,	/* iManufacturer */
    0x02,	/* iProduct */
    0x03,	/* SerialNumber */
    0x01	/* bNumConfigs */
};

const uint8_t cfgDescriptor[] = {	/* Configuration */
    0x09,	/* CbLength */
    DESC_CONFIG,	/* CbDescritprType */
    0x43, 0x00,	/* CwTotalLength 2 EP + Control */
    0x02,	/* CbNumInterfaces */
    0x01,	/* CbConfigurationValue */
    0x00,	/* CiConfiguration */
    0xC0,	/* CbmAttributes */
    0x64,	/* CMaxPower */

    /* Communication Class Interface Descriptor Requirement */
    0x09,	/* bLength */
    DESC_INTERFACE,	/* bDescriptorType */
    0x00,	/* bInterfaceNumber #0 */
    0x00,	/* bAlternateSetting #0 */
    0x01,	/* bNumEndpoints 1 */
    0x02,	/* bInterfaceClass */
    0x02,	/* bInterfaceSubclass (ACM) */
    0x00,	/* bInterfaceProtocol */
    0x00,	/* iInterface */

    /* Header Function Descriptor */
    0x05,	/* bFunction Length */
    0x24,	/* bDescriptor type: CS_INTERFACE */
    0x00,	/* bDescriptor subtype: Header Func Desc */
    0x10, 0x01,	/* bcdCDC:1.1 */

    /* ACM Functional Descriptor */
    0x04,	/* bFunctionLength */
    0x24,	/* bDescriptor Type: CS_INTERFACE */
    0x02,	/* bDescriptor Subtype: ACM Func Desc */
    0x02,	/* bmCapabilities */

    /* Union Functional Descriptor */
    0x05,	/* bFunctionLength */
    0x24,	/* bDescriptorType: CS_INTERFACE */
    0x06,	/* bDescriptorSubtype: Union Func Desc */
    0x00,	/* bMasterInterface: Communication Class Interface */
    0x01,	/* bSlaveInterface0: Data Class Interface */

    /* Call Management Functional Descriptor */
    0x05,	/* bFunctionLength */
    0x24,	/* bDescriptorType: CS_INTERFACE */
    0x01,	/* bDescirptorSubtype: Call Management Function */
    0x00,	/* bmCapabilities: Self call management */
    0x01,	/* bDataInterface: Data Class Interface */

    /* Endpoint descriptor */
    0x07,	/* bLength */
    0x05,	/* bDescriptorType */
    0x80 | CDC_EP_NOTICE,	/* bEndpointAddress, Endpoint 03 - IN */
    0x03,	/* bmAttributes	INT */
    0x08, 0x00,	/* wMaxPacketSize */
    0xFF,	/* bInterval */

    /* Data Class Interface Descriptor Requirement */
    0x09,	/* bLength */
    0x04,	/* bDescriptorType */
    0x01,	/* bInterfaceNumber */
    0x00,	/* bAlternateSetting */
    0x02,	/* bNumEndpoints */
    0x0a,	/* bInterfaceClass */
    0x00,	/* bInterfaceSubclass */
    0x00,	/* bInterfaceProtocol */
    0x00,	/* iInterface */

    /* First alternate setting. Endpoint 1 descriptor */
    0x07,	/* bLength */
    0x05,	/* bDescriptorType */
    0x00 | CDC_EP_OUT,	/* bEndpointAddress, Endpoint 01 - OUT */
    0x02,	/* bmAttibutes BULK */
    EP_SIZE_BULK, 00,	/* wMaxPacketSize */
    0x00,	/* bInterval */

    /* Endpoint 2 descriptor */
    0x07,	/* bLength */
    0x05,	/* bDescriptorType */
    0x80 | CDC_EP_IN,	/* bEndpointAddress, Endpoint 02 - IN */
    0x02,	/* bmAttibutes BULK */
    EP_SIZE_BULK, 00,	/* wMaxPacketSize */
    0x00,	/* bInterval */
};

const uint8_t languageDescriptor[] = {
    0x04,	/* length */
    0x03,	/* String descriptor */
    0x09, 0x04,	/* English, US */
};

const uint8_t manufacturerDescriptor[] = {
    12,		/* length */
    0x03,	/* String descriptor */
    'A', 0x00, 't', 0x00, 'm', 0x00, 'e', 0x00, 'l', 0x00,
};

const uint8_t productDescriptor[] = {
    0x12,	/* length */
    0x03,	/* String descriptor */
    'V', 0x00, 'i', 0x00, 'r', 0x00, 't', 0x00,
    'u', 0x00, 'a', 0x00, 'l', 0x00, ' ', 0x00,
    'C', 0x00, 'O', 0x00, 'M', 0x00, ' ', 0x00,
    'p', 0x00, 'o', 0x00, 'r', 0x00, 't', 0x00,
};

const uint8_t serialNumberDescriptor[] = {
    0x8,	/* length */
    0x03,	/* String descriptor */
    '3', 0x00, '2', 0x00, '1', 0x00,
};

const struct strdesc {
    const uint8_t *string;
    int length;
} stringDescriptors[] = {
    { languageDescriptor, sizeof(languageDescriptor) },
    { manufacturerDescriptor, sizeof(manufacturerDescriptor), },
    { productDescriptor, sizeof(productDescriptor) },
    { serialNumberDescriptor, sizeof(serialNumberDescriptor) },
};

struct cdc_line_coding cdc_line = {
    115200,	/* Baud rate */
    0,		/* 1 stop bit */
    0,		/* no parity */
    8,		/* data bits */
};

static char bannar[80];

void udp_read_fifo(Udp *pUDP, int epno, struct ep_desc *pEpDesc, int nbyte)
{
  if (nbyte > pEpDesc->remain) {
    nbyte = pEpDesc->remain;
  }
  pEpDesc->remain -= nbyte;
  pEpDesc->done += nbyte;
  while (nbyte > 0) {
    *(pEpDesc->pdata) = pUDP->UDP_FDR[epno];
    pEpDesc->pdata++;
    nbyte--;
  }
}

void udp_fill_fifo(Udp *pUDP, int epno, struct ep_desc *pEpDesc)
{
  int size;

  size = pEpDesc->epsize;
  if (size > pEpDesc->remain)
    size = pEpDesc->remain;
  pEpDesc->remain -= size;
  pEpDesc->filled += size;

  /* Write packet into FIFO */
  while (size > 0) {
    pUDP->UDP_FDR[epno] = *(pEpDesc->pdata);
    pEpDesc->pdata++;
    size--;
  }
}

void usb_ep_read(Udp *pUDP, int epno, uint8_t *pdata, int len, void (*handler)())
{
  struct ep_desc *pEpDesc = &epdesctab[epno];

  if (pEpDesc->epstate != EP_IDLE)
    return;
  pUDP->UDP_IDR = (1 << epno);
  pEpDesc->epstate = EP_RECEIVING;
  pEpDesc->pdata = pdata;
  pEpDesc->remain = len;
  pEpDesc->filled = 0;
  pEpDesc->done = 0;
  pEpDesc->callback = handler;

  pUDP->UDP_IER = (1 << epno);
}

void usb_ep_write(Udp *pUDP, int epno, uint8_t *pdata, int len)
{
  struct ep_desc *pEpDesc = &epdesctab[epno];

  if (pEpDesc->epstate != EP_IDLE) {
    return;
  }
  pUDP->UDP_IDR = (1 << epno);
  pEpDesc->pdata = pdata;
  pEpDesc->remain = len;
  pEpDesc->filled = 0;
  pEpDesc->done = 0;
  while (pUDP->UDP_CSR[epno] & UDP_CSR_TXPKTRDY)
    ;
  /* Start sending first packet */
  pEpDesc->epstate = EP_SENDING;
  udp_fill_fifo(pUDP, epno, pEpDesc);
  SET_CSR(epno, UDP_CSR_TXPKTRDY);

  /* If necessary, fill second packet */
  if (pEpDesc->dualbank && (pEpDesc->remain > 0)) {
    udp_fill_fifo(pUDP, epno, pEpDesc);
  }
  pUDP->UDP_IER = (1 << epno);
}

void process_ep0(Udp *pUDP)
{
  uint8_t *wp;
  uint32_t csr, rxbytes, i;
  struct ep_desc *pEpDesc = &epdesctab[0];

  csr = pUDP->UDP_CSR[0];
  rxbytes = (csr >> 16) & 0x7f;

  if (csr & UDP_CSR_RXSETUP) {	/* Setup packet */
    if (pEpDesc->epstate != EP_IDLE)
      return;
    wp = (uint8_t *)&req_pkt;
    for (i = 0; i < rxbytes; i++) {
      *wp++ = pUDP->UDP_FDR[0];
    }
    if (req_pkt.bmRequestType & DIR_DEV_TO_HOST) {	/* IN direction */
      SET_CSR(0, UDP_CSR_DIR);
    } else {
      CLEAR_CSR(0, UDP_CSR_DIR);
    }
    /* Tell hims that we have read the FIFO */
    CLEAR_CSR(0, UDP_CSR_RXSETUP);

    if (rxbytes > 0) {
      pEpDesc->callback = usb_process_setup;
      pEpDesc->epstate = EP_DONE;
    }
  }
  if (csr & UDP_CSR_TXCOMP) {
    if (pEpDesc->epstate == EP_SENDING) {
      if (pEpDesc->filled < pEpDesc->epsize) {
        /* last packet has sent */
        pEpDesc->done += pEpDesc->filled;
        pEpDesc->filled = 0;
        pEpDesc->epstate = EP_IDLE;
      } else {
        /* Transfer remaining data */
        pEpDesc->done += pEpDesc->epsize;
        pEpDesc->filled -= pEpDesc->epsize;
        udp_fill_fifo(pUDP, 0, pEpDesc);
        SET_CSR(0, UDP_CSR_TXPKTRDY);
      }
    }
    CLEAR_CSR(0, UDP_CSR_TXCOMP);
    if (addr_flag) {
        pUDP->UDP_FADDR = (UDP_FADDR_FEN | addr_flag);
        pUDP->UDP_GLB_STAT = UDP_GLB_STAT_FADDEN;
        addr_flag = 0;
    }
  }
  if (csr & UDP_CSR_RX_DATA_BK0) {
    if (pEpDesc->epstate == EP_RECEIVING) {
      udp_read_fifo(pUDP, 0, pEpDesc, rxbytes);
      CLEAR_CSR(0, UDP_CSR_RX_DATA_BK0);
      if (pEpDesc->remain == 0)
    	pEpDesc->epstate = EP_DONE;
    } else {
      if (rxbytes == 0) {	/* ACK has received */
    	CLEAR_CSR(0, UDP_CSR_RX_DATA_BK0);
    	pEpDesc->epstate = EP_IDLE;
      } else if (csr & UDP_CSR_FORCESTALL) {
    	CLEAR_CSR(0, UDP_CSR_RX_DATA_BK0);
      } else {	/* NAK the data */
    	pUDP->UDP_IDR = (1 << 0);
      }
    }
  }
  if (csr & UDP_CSR_STALLSENTISOERROR) {
    CLEAR_CSR(0, UDP_CSR_STALLSENTISOERROR | UDP_CSR_FORCESTALL);
  }
}

void process_ep(Udp *pUDP, int epno)
{
  uint32_t csr, nb;
  struct ep_desc *pEpDesc = &epdesctab[epno];

  csr = pUDP->UDP_CSR[epno];

  if (csr & (UDP_CSR_RX_DATA_BK0|UDP_CSR_RX_DATA_BK1)) {
    nb = (csr >> 16) & 0x7ff;
    if (pEpDesc->epstate == EP_RECEIVING) {
      udp_read_fifo(pUDP, epno, pEpDesc, nb);
      if (pEpDesc->cbank == UDP_CSR_RX_DATA_BK0) {
        if (pEpDesc->dualbank) {
          CLEAR_CSR(epno, UDP_CSR_RX_DATA_BK0);
          pEpDesc->cbank = UDP_CSR_RX_DATA_BK1;
        }
      } else {
        CLEAR_CSR(epno, UDP_CSR_RX_DATA_BK1);
        pEpDesc->cbank = UDP_CSR_RX_DATA_BK0;
      }
      if ((pEpDesc->remain == 0) || (nb < EP_SIZE_BULK)) {
        pEpDesc->epstate = EP_DONE;
        pUDP->UDP_IDR = (1 << epno);
      }
    } else {
      if (csr & UDP_CSR_FORCESTALL) {
        if (pEpDesc->cbank == UDP_CSR_RX_DATA_BK0) {
          CLEAR_CSR(epno, UDP_CSR_RX_DATA_BK0);
          if (pEpDesc->dualbank) {
            pEpDesc->cbank = UDP_CSR_RX_DATA_BK1;
          }
        } else {
          CLEAR_CSR(epno, UDP_CSR_RX_DATA_BK1);
          pEpDesc->cbank = UDP_CSR_RX_DATA_BK0;
        }
      } else {
        /* NAK */
        pUDP->UDP_IDR = (1 << epno);
      }
    }
  } else if (csr & UDP_CSR_TXCOMP) {
    if (pEpDesc->epstate == EP_SENDING) {
      if ((pEpDesc->filled <= pEpDesc->epsize) &&
         (pEpDesc->remain == 0)) {
        /* All data has been transfered */
        pEpDesc->done += pEpDesc->filled;
        pEpDesc->epstate = EP_IDLE;
        pUDP->UDP_IDR = (1 << epno);
        CLEAR_CSR(epno, UDP_CSR_TXCOMP);
      } else {
        /* Transfer remaining data */
        pEpDesc->done += pEpDesc->epsize;
        pEpDesc->filled -= pEpDesc->epsize;
        if (pEpDesc->dualbank) {
          /* Double buffering */
          SET_CSR(epno, UDP_CSR_TXPKTRDY);
          CLEAR_CSR(epno, UDP_CSR_TXCOMP);
          udp_fill_fifo(pUDP, 0, pEpDesc);
        } else {
          CLEAR_CSR(epno, UDP_CSR_TXCOMP);
        }
      }
    } else {
      CLEAR_CSR(epno, UDP_CSR_TXCOMP);
    }
  }
  if (csr & UDP_CSR_STALLSENTISOERROR) {
    CLEAR_CSR(epno, UDP_CSR_STALLSENTISOERROR | UDP_CSR_FORCESTALL);
  }
}

/* UDP interrupt handler */
void UDP_IRQHandler()
{
  Udp *pUDP = UDP;
  uint32_t isr;

  isr = pUDP->UDP_ISR;
  if (isr & UDP_ISR_ENDBUSRES) {	/* BUS reset */
    pUDP->UDP_ICR = UDP_ICR_ENDBUSRES;
    /* Reset all end points */
    pUDP->UDP_RST_EP = ~0;
    pUDP->UDP_RST_EP = 0;
    /* Enable the function */
    pUDP->UDP_FADDR = UDP_FADDR_FEN;
    addr_flag = 0;
    /* Configure EP0 */
    pUDP->UDP_CSR[EP_CTRL] = (UDP_CSR_EPEDS | UDP_CSR_EPTYPE_CTRL);
    usb.currentConfiguration = 0;
    usb.currentConnection = 0;
    /* IMR is reset with BUSRES. We need to initialize IER */
    pUDP->UDP_IDR = ~0;
    pUDP->UDP_IER = (1 << 0);
    pUDP->UDP_TXVC &= ~UDP_TXVC_TXVDIS;
  } else {
    if (isr & (1 << 0)) {
      process_ep0(pUDP);
    }
    if (isr & (1 << CDC_EP_OUT)) {
      process_ep(pUDP, CDC_EP_OUT);
    }
    if (isr & (1 << CDC_EP_IN)) {
      process_ep(pUDP, CDC_EP_IN);
    }
    if (isr & (1 << CDC_EP_NOTICE)) {
      process_ep(pUDP, CDC_EP_NOTICE);
    }
  }
  pUDP->UDP_ICR = isr;
}

void udp_init(void)
{
  Udp *pUDP = UDP;

  pUDP->UDP_TXVC = 0;		/* Enable tranceiver */
  pUDP->UDP_ICR = ~0;		/* Clear UDP_ISR */
  pUDP->UDP_TXVC = UDP_TXVC_PUON;
  NVIC_EnableIRQ(UDP_IRQn);	/* Enable UDP interrupt */
}

void udp_disable(void)
{
  Udp *pUDP = UDP;

  pUDP->UDP_TXVC = UDP_TXVC_TXVDIS;		/* Disable tranceiver */
}

/* Send Zero Length Packet */
void usb_send_zlp(Udp *pUDP)
{
  SET_CSR(EP_CTRL, UDP_CSR_TXPKTRDY);
}

/* Send Stall signal to the host */
void usb_send_stall(Udp *pUDP, int epno)
{
  SET_CSR(epno, UDP_CSR_FORCESTALL);
}

void linecoding_callback(Udp *pUDP, struct ep_desc *pEpDesc)
{
  usb_send_zlp(pUDP);
  pEpDesc->epstate = EP_IDLE;
}

uint16_t serialState;

struct usb_notification {
  struct usb_pkt pkt;
  uint8_t pdata[2];
} __attribute__((packed));

static void usb_process_setup(Udp *pUDP, struct ep_desc *pEpDesc)
{
  struct usb_pkt *up = &req_pkt;
  uint16_t code;
  uint8_t sval;

  pEpDesc->epstate = EP_IDLE;

  code = (up->bRequest << 8) | up->bmRequestType;
  switch (code) {
  case STD_GET_DESCRIPTOR:
    sval = up->wValue & 0xff;

    switch (up->wValue >> 8) {
    case DESC_DEVICE:	/* Device Descriptor */
      usb_ep_write(pUDP, 0, (uint8_t *)devDescriptor, MIN(sizeof(devDescriptor), up->wLength));
      break;
    case DESC_CONFIG:	/* Config Descriptor */
      usb_ep_write(pUDP, 0, (uint8_t *)cfgDescriptor, MIN(sizeof(cfgDescriptor), up->wLength));
      break;
    case DESC_STRING:
      if (sval < 4) {
        const struct strdesc *pstr = &stringDescriptors[sval];

        usb_ep_write(pUDP, 0, (uint8_t *)pstr->string, pstr->length);
      } else {
        usb_send_stall(pUDP, 0);
      }
      break;
    default:
      usb_send_stall(pUDP, 0);
      break;
    }
    break;
  case STD_SET_ADDRESS:
    usb_send_zlp(pUDP);
    addr_flag = up->wValue;
    break;
  case STD_SET_CONFIGURATION:
    usb.currentConfiguration = up->wValue;
    usb_send_zlp(pUDP);
    if (up->wValue) {
      /* Enable endpoints */
      pUDP->UDP_CSR[CDC_EP_OUT] = UDP_CSR_EPEDS | UDP_CSR_EPTYPE_BULK_OUT;
      pUDP->UDP_CSR[CDC_EP_IN] = UDP_CSR_EPEDS | UDP_CSR_EPTYPE_BULK_IN;
      pUDP->UDP_CSR[CDC_EP_NOTICE] = UDP_CSR_EPEDS | UDP_CSR_EPTYPE_INT_IN;
      pUDP->UDP_IER = (1 << CDC_EP_OUT) | (1 << CDC_EP_IN) | (1 << CDC_EP_NOTICE);
      pUDP->UDP_GLB_STAT = UDP_GLB_STAT_CONFG;
    } else {
      /* Disable endpoints */
      pUDP->UDP_CSR[CDC_EP_OUT] = pUDP->UDP_CSR[CDC_EP_IN] =
      pUDP->UDP_CSR[CDC_EP_NOTICE] = 0;
      pUDP->UDP_IDR = (1 << CDC_EP_OUT) | (1 << CDC_EP_IN) | (1 << CDC_EP_NOTICE);
      pUDP->UDP_GLB_STAT = UDP_GLB_STAT_FADDEN;
    }
    epdesctab[CDC_EP_IN].cbank = UDP_CSR_RX_DATA_BK0;
    epdesctab[CDC_EP_OUT].cbank = UDP_CSR_RX_DATA_BK0;
    break;
  case CDC_SET_LINECODING:
    usb_ep_read(pUDP, 0, (uint8_t *)&cdc_line, sizeof(cdc_line), linecoding_callback);
    break;
  case CDC_GET_LINECODING:
    usb_ep_write(pUDP, 0, (uint8_t *)&cdc_line, sizeof(cdc_line));
    break;
  case CDC_SET_CTRL_LINE_STATE:
    if (usb.currentConnection != (up->wValue & 1)) {
      if (usb.currentConnection == 0) {
        snprintf(bannar, sizeof(bannar), "]]] OV7670 %s [[[\r\n", OPENING_TITLE);
        usb_ep_write(pUDP, CDC_EP_IN, (uint8_t *)bannar, strlen(bannar));
      }
    }
    usb.currentConnection = up->wValue & 1;
    usb_send_zlp(pUDP);
    break;
  default:
    usb_send_stall(pUDP, 0);
    break;
  }
  pUDP->UDP_IER = (1 << 0);
}

void usb_task()
{
  struct ep_desc *pEpDesc = epdesctab;
  int i;

  for (i = 0; i < EPDESCTAB_SIZE; i++) {
    if (pEpDesc->epstate == EP_DONE) {
      if (pEpDesc->callback) {
        (pEpDesc->callback)(UDP, pEpDesc);
      }
    }
    pEpDesc++;
  }
}

void usb_start_xmit(char *bp, int size)
{
  usb_ep_write(UDP, CDC_EP_IN, (uint8_t *)bp, size);
}

int cdc_write(uint8_t *bp, int size)
{
  if (usb.currentConnection) {
    usb_ep_write(UDP, CDC_EP_IN, bp, size);
  }
  return size;
}

void recv_callback(Udp *pUDP, struct ep_desc *pEpDesc)
{
  pEpDesc->epstate = EP_IDLE;
}

void usb_start_recv(char *bp, int size)
{
  usb_ep_read(UDP, CDC_EP_OUT, (uint8_t *)bp, size, recv_callback);
}
